/* <domex>/src/sd3/sd/FragmentIdentifier has several values in the function "updateCosineParams"
 Ok.  Same file, "mp3FrameLengthFromHeader" has my code for the frame length analyzer.
 */

#include "config.h"
#include "block_scan.h"
#include "bloom.h"
#include <openssl/md5.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <fcntl.h>

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <set>
#include <vector>
#include <algorithm>

#include "discriminator.h"
#include "block_scan.h"
#include "aftimer.h"
#include "utils.h"

extern int opt_debug;

discriminator_list discriminators;

string itos(int i){
    char buf[1024];
    snprintf(buf,sizeof(buf),"%d",i);
    return string(buf);
}


void discriminator_list::init()
{
    if (discriminators.size()==0){
	discriminators.push_back(new jpeg_discriminator());
	discriminators.push_back(new null_discriminator());
	discriminators.push_back(new entropy_discriminator());
	discriminators.push_back(new huffman_discriminator());
	discriminators.push_back(new pdf_discriminator());
    }
}

static bool count_ascend_fun(const char_count &a,const char_count &b)
{
    return a.count > b.count;
}

static inline double square(double x){
    return x*x;
}

votelist_t null_discriminator::identify(const sdbuf &buf){
    votelist_t ret;
    if(buf.null_block()){
	ret.push_back(vote(buf.block_number,"NULL","",1.0,0.0,10));
    }
    return ret;
}

votelist_t jpeg_discriminator::identify(const sdbuf &buf)
{
    votelist_t ret;
    if(buf[0]==0xff && buf[1]==0xd8 && buf[2]==0xff && (buf[3]==0xe0||buf[3]==0xe1)){
	ret.push_back(vote(buf.block_number,"JPEG","JPEG Header",1.0,0.0,9));
    }
    if(buf.h.unique_unigrams()>JPEG_HE){
	string ff00("\xff\x00");
	if(buf.h.ngram_count(ff00)>JPEG_LN){
	    ret.push_back(vote(buf.block_number,"JPEG","JPEG FF00 Region",0.998,0.015,9));
	}
    }
    return ret;
}


votelist_t entropy_discriminator::identify(const sdbuf &buf){
    votelist_t ret;
    if(buf.h.unique_unigrams()>ENTROPY_HIGH){
	ret.push_back(vote(buf.block_number,"ENTROPY HIGH","High entropy",1.0,0.0,1));
    }
    if(buf.h.unique_unigrams()<ENTROPY_LOW){
	ret.push_back(vote(buf.block_number,"ENTROPY LOW","Low entropy",1.0,0.0,1));
    }
    /* Scan for non-ascii */
    int non_ascii_count = 0;
    for(int i=0;i<256;i++){
	if((i<' ' && i!='\r' && i!='\t' && i!='\n' ) ||
	   (i>127)){
	    non_ascii_count += buf.h[i];
	}
    }
    if(non_ascii_count < MAX_NON_ASCII){
	ret.push_back(vote(buf.block_number,"ASCII","printable 7-bit ascii",1.0,0.0,2));
    }
    return ret;
}


votelist_t huffman_discriminator::identify(const sdbuf &buf)
{
    const int HE=230;	 // guess value
    const int VL=255;	 // according to Alex, this is the right value
    float MCV=0.9990;	 // according to Alex, this is the right value
    votelist_t ret;
    sdbuf autobuf(buf.block_size);		// for the autocorrelation
    for(size_t i=0;i<buf.block_size;i++){
	autobuf[i] = buf[i] + buf[(i+1)%buf.block_size];
    }

    if(buf.h.unique_unigrams()<HE) return ret; // not random enough

    /* Compute the histogram of each */
    autobuf.calc_block_histogram();
    vector<char_count> buf_counts = buf.h.get_char_counts();
    vector<char_count> autobuf_counts = autobuf.h.get_char_counts();

    /* Sort the histograms */
    sort(buf_counts.begin(),buf_counts.end(),count_ascend_fun);
    sort(autobuf_counts.begin(),autobuf_counts.end(),count_ascend_fun);

    /* Compute the dot product for the first VL elements in the histogram */
    double dotproduct = 0.0;
    for(int i=0;i<VL;i++){
	dotproduct += buf_counts[i].count*autobuf_counts[i].count;
    }

    double mag_squared1 = 0.0;
    double mag_squared2 = 0.0;
    for(int i=0;i<VL;i++){
	mag_squared1 += square(buf_counts[i].count);
	mag_squared2 += square(autobuf_counts[i].count);
    }
    double mag1 = sqrt(mag_squared1);
    double mag2 = sqrt(mag_squared2);
    double cosinesim = dotproduct / (mag1 * mag2);
    if(opt_debug>1) printf("cosinesim=%g\n",cosinesim);
    if(cosinesim<MCV){
	ret.push_back(vote(buf.block_number,"HUFFMAN","Cosine test",0.50,0.03,5));
    }

    return ret;
}

votelist_t pdf_discriminator::identify(const sdbuf &buf)
{
    votelist_t ret;
    if(reg1==0){
	reg1 = new myregex("[0-9]{10} [0-9]{5} n(\r\n|\r|\n)[0-9]{10} [0-9]{5} n(\r\n|\r|\n",0);
    }
    /* We can't use a straight regular expression because nulls in the target
     * would cause regexec to abort...
     */
    for(size_t i=0;i<buf.block_size-43;i++){
	if(
	   (buf[i]=='n' && buf[i+1]=='\r' && buf[i+2]=='\n' && buf[i+3]=='0' && reg1->search(buf.substr(i+3,40))) ||
	   (buf[i]=='n' && buf[i+1]=='\r' && buf[i+2]=='0'  && reg1->search(buf.substr(i+2,40))) ||
	   (buf[i]=='n' && buf[i+1]=='\n' && buf[i+2]=='0'  && reg1->search(buf.substr(i+2,40)))){
	    ret.push_back(vote(buf.block_number,"PDF","characteristic offset list",1.0,0.0,5));
	       break;
	}
    }
    return ret;
}


bloom_discriminator *bloom_discriminator::new_bloom(string fn)
{
    NSRLBloom *bf = new NSRLBloom();
    if(bf->open(fn.c_str(),O_RDONLY)){
	delete bf;
	return 0;
    }
    bloom_discriminator *bd = new bloom_discriminator();
    bd->fname = fn;
    bd->bf = bf;
    return bd;
}

votelist_t bloom_discriminator::identify(const sdbuf &buf)
{
    u_char md5hash[16];
    votelist_t ret;

    for(size_t offset=0;offset<buf.block_size &&  buf.block_size + offset < buf.buflen; offset+=512){
	MD5(buf.buf+offset,buf.block_size,md5hash);
	if(bf->query(md5hash)){
	    char md5hex[33];
	    nsrl_hexbuf(md5hex,sizeof(md5hex),md5hash,16,0);
	    ret.push_back(vote(buf.block_number,"BLOOM",
			       "filter "+fname+" block offset "+itos(offset)+" "+string(md5hex),
			       1.0,0.0,10));
	    break;
	}
    }
    return ret;
}
